"""
Affine transforms in three dimensions.
"""

from __future__ import annotations

from typing import TypeVar

import os
import torch
import voxel as vx


T = TypeVar('T', bound='AffineMatrix')


class AffineMatrix:
    """
    Affine matrix (4x4) transform for a 3D coordinate system.
    """

    def __init__(self,
        data: torch.Tensor | None = None,
        device: torch.device | None = None) -> None:
        """
        Args:
            data (Tensor, optional): A 3x3, 3x4, or 4x4 tensor. Default: if
                None, the matrix is initialized with the identitiy.
            device (device, optional): Device of the constructed matrix.
        """
        vx.caching.init_property_cache(self)

        if data is None:
            data = torch.eye(4, device=device)
        elif isinstance(data, AffineMatrix):
            data = data.tensor

        data = torch.as_tensor(data, device=device).float()

        if data.shape == (3, 3):
            row = torch.zeros((3, 1), dtype=data.dtype, device=data.device)
            data = torch.cat((data, row), dim=1)

        if data.shape == (3, 4):
            row = torch.tensor([[0, 0, 0, 1]], dtype=data.dtype, device=data.device)
            data = torch.cat((data, row), dim=0)
        elif data.shape != (4, 4):
            raise ValueError('Input matrix must be 3x3, 3x4, or 4x4.')

        self._tensor = data

    @property
    def tensor(self) -> torch.Tensor:
        """
        Tensor data of shape (4, 4).
        """
        return self._tensor

    def __getitem__(self, indexing) -> torch.Tensor:
        return self.tensor[indexing]

    def __setitem__(self, indexing, item):
        self.tensor[indexing] = item

    def __repr__(self) -> str:
        name = self.__class__.__name__
        tensor_str = str(self.tensor.detach().cpu().numpy())
        tensor_str = tensor_str.replace('\n', f'\n{" " * (len(name) + 1)}')
        return f'{name}({tensor_str})'

    def __matmul__(self, other: AffineMatrix | torch.Tensor) -> AffineMatrix | torch.Tensor:
        isaffine = isinstance(other, AffineMatrix) or \
                    (torch.is_tensor(other) and other.shape == (4, 4))
        if isaffine:
            other = AffineMatrix(other).tensor
        result = self.tensor.to(other.device) @ other.type(self.tensor.dtype)
        return AffineMatrix(result) if isaffine else result

    def _from_tensor_with_new_properties(self: T, tensor: torch.Tensor) -> T:
        """
        Base class utility function that creates a new object instance, with a
        new matrix tensor, but the same metadata. This should be reimplemented
        for subclasses. This function should be called in scenarios only when the
        matrix has new properties (e.g. device or data type), not new values.
        """
        return self.__class__(tensor)

    def save(self, filename: os.PathLike, fmt: str = None) -> None:
        """
        Save the affine matrix to a file.

        Args:
            filename (PathLike): The path to the file to save.
            fmt (str, optional): The format of the file. If None, the format is
                determined by the file extension.
        """
        vx.save_affine(self, filename, fmt=fmt)

    def detach(self: T) -> T:
        """
        Detach the matrix tensor from the current computational graph.

        Returns:
            A new affine with the detached matrix tensor.
        """
        return self._from_tensor_with_new_properties(self.tensor.detach())

    def to(self: T, device: torch.Device) -> T:
        """
        Move the matrix tensor to a device.

        Args:
            device (Device): The target device.

        Returns:
            A new affine with the matrix tensor on the target device.
        """
        return self._from_tensor_with_new_properties(self.tensor.to(device))

    def cuda(self: T) -> T:
        """
        Move the matrix tensor to the GPU.

        Returns:
            A new affine with the matrix tensor on the GPU.
        """
        return self._from_tensor_with_new_properties(self.tensor.cuda())

    def cpu(self: T) -> T:
        """
        Move the matrix tensor to the CPU.

        Returns:
            A new affine with the matrix tensor on the CPU.
        """
        return self._from_tensor_with_new_properties(self.tensor.cpu())

    def inverse(self) -> AffineMatrix:
        """
        Invert the matrix.

        Returns:
            AffineMatrix: Inverted affine matrix.
        """
        return AffineMatrix(self.tensor.inverse())

    def transform(self, coords: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix transformation to a set of 3D coordinates.

        Args:
            coordinates (Tensor): A tensor of coordinates with shape (..., 3).

        Returns:
            Tensor: Transformed coordinates with the same shape as the input.
        """
        if coords.shape[-1] != 3:
            raise ValueError('Coordinates must have a last dimension of size 3.')

        # reshape to a 2D tensor for the transformation
        coords_reshaped = coords.reshape(-1, 3)

        # convert to homogeneous coordinates
        ones = torch.ones((coords_reshaped.shape[0], 1), dtype=coords.dtype, device=coords.device)
        coords_homogeneous = torch.cat((coords_reshaped, ones), dim=1)

        # apply the transformation, convert back to cartesian, and reshape
        transformed_coords = coords_homogeneous @ self.tensor.T.to(coords.device)
        return transformed_coords[:, :3].reshape(coords.shape)


class AffineVolumeTransform(AffineMatrix):
    """
    Affine transform matrix in world or voxel space that contains metadata
    about the source and target acquistion geometry.
    """

    def __init__(self,
        data: torch.Tensor | AffineMatrix,
        space: vx.Space,
        source: vx.AcquisitionGeometry | vx.Volume,
        target: vx.AcquisitionGeometry | vx.Volume) -> None:
        """
        Args:
            data (Tensor | AffineMatrix): A 3x3, 3x4, or 4x4 tensor matrix.
            space (Space): The coordinate space of the transform.
            source (AcquisitionGeometry | Volume): The source acquisition geometry.
            target (AcquisitionGeometry | Volume): The target acquisition geometry.
        """
        super().__init__(data)
        self._source = vx.acquisition.cast_acquisition_geometry(source)
        self._target = vx.acquisition.cast_acquisition_geometry(target)
        self._space = vx.Space(space)
    
    @property
    def space(self) -> vx.Space:
        """
        Coordinate space of the transform.
        """
        return self._space
    
    @property
    def source(self) -> vx.AcquisitionGeometry:
        """
        Source acquisition geometry.
        """
        return self._source
    
    @property
    def target(self) -> vx.AcquisitionGeometry:
        """
        Target acquisition geometry.
        """
        return self._target

    def inverse(self) -> AffineVolumeTransform:
        """
        Invert the transform.

        Returns:
            AffineVolumeTransform: Inverted affine transform.
        """
        inverse = self.tensor.inverse()
        return AffineVolumeTransform(inverse, self.space, self.target, self.source)

    def convert(self,
        space: vx.Space | None = None,
        source: vx.AcquisitionGeometry | vx.Volume | None = None,
        target: vx.AcquisitionGeometry | vx.Volume | None = None) -> AffineVolumeTransform:
        """
        Convert transform for a new set of coordinate space, source, or target.

        Parameters
        ----------
        space (Space, optional): Desired coordinate space.
        source (AcquisitionGeometry or Volume, optional): Desired source geometry.
        target (AcquisitionGeometry or Volume, optional): Desired target geometry.

        Returns:
            AffineVolumeTransform: Converted affine transform.
        """
        
        # check if the desired space is the same as the embedded space
        space = self.space if space is None else vx.Space(space)
        same_space = space == self.space

        # check if the desired source is the same as the embedded source
        if source is None:
            source = self.source
            same_source = True
        else:
            source = vx.acquisition.cast_acquisition_geometry(source)
            same_source = torch.allclose(source.tensor, self.source.tensor, atol=1e-4, rtol=0)

        # check if the desired target is the same as the embedded target
        if target is None:
            target = self.target
            same_target = True
        else:
            target = vx.acquisition.cast_acquisition_geometry(target)
            same_target = torch.allclose(target.tensor, self.target.tensor, atol=1e-4, rtol=0)

        # return self if no changes are needed
        if all((same_space, same_source, same_target)):
            return self

        if same_source and same_target:
            # just a simple conversion of transform coordinate space, without
            # changing source and target information
            a = self.target if space == 'world' else self.target.inverse()
            b = self.source if space == 'voxel' else self.source.inverse()
            affine = a @ self @ b
        else:
            # if source and target info is changing, we need to recompute the
            # transform by first converting it to world-space
            affine = self if self.space == 'world' else self.target @ self @ self.source.inverse()
            # then back into the desired coordinate space
            if space == 'voxel':
                affine = target.inverse() @ affine @ source

        return AffineVolumeTransform(affine, space, source, target)


def translation_matrix(translation: torch.Tensor) -> AffineMatrix:
    """
    Compute a 3D translation matrix from translation vector.

    Args:
        translation (Tensor): Translation vector.
    
    Returns:
        AffineMatrix: Translation affine matrix.
    """
    if translation.shape != (3,):
        raise ValueError('Translation vector must have a shape of (3,).')
    matrix = torch.eye(4, dtype=torch.float64, device=translation.device)
    matrix[:3, 3] = translation
    return AffineMatrix(matrix)


def angles_to_rotation_matrix(
    rotation: torch.Tensor,
    degrees: bool = True) -> AffineMatrix:
    """
    Compute a 3D rotation matrix from rotation angles.

    Args:
        rotation (Tensor): Rotation angles. If `degrees` is True, the
            angles are in degrees, otherwise they are in radians.
        degrees (bool, optional): Whether the angles are defined as degrees or,
            alternatively, as radians.

    Returns:
        AffineMatrix: Rotation affine matrix.
    """
    if degrees:
        rotation = torch.deg2rad(rotation)

    c, s = torch.cos(rotation[0]), torch.sin(rotation[0])
    rx = torch.tensor([[1, 0, 0], [0, c, s], [0, -s, c]], dtype=torch.float64)
    c, s = torch.cos(rotation[1]), torch.sin(rotation[1])
    ry = torch.tensor([[c, 0, s], [0, 1, 0], [-s, 0, c]], dtype=torch.float64)
    c, s = torch.cos(rotation[2]), torch.sin(rotation[2])
    rz = torch.tensor([[c, s, 0], [-s, c, 0], [0, 0, 1]], dtype=torch.float64)
    matrix = rx @ ry @ rz

    return AffineMatrix(matrix.to(rotation.device))


def compose_affine(
    translation : torch.Tensor = None,
    rotation : torch.Tensor = None,
    scale : torch.Tensor = None,
    shear : torch.Tensor = None,
    degrees : bool = True,
    device : torch.device = None) -> AffineMatrix:
    """
    Composes an affine matrix from a set of translation, rotation, scale,
    and shear transform components.

    Args:
        translation (Tensor, optional): Translation vector.
        rotation (Tensor, optional): Rotation angles.
        scale (Tensor, optional): Scaling factors.
        shear (Tensor, optional): Shearing factors.
        degrees (bool, optional): Whether the rotation angles are in degrees.
        device (device, optional): Device of the generated matrix.

    Returns:
        AffineMatrix: Composed affine matrix.
    """
    # check translation
    translation = torch.zeros(3) if translation is None else torch.as_tensor(translation)
    if len(translation) != 3:
        raise ValueError(f'translation must be of shape (3,)')

    # check rotation angles
    rotation = torch.zeros(3) if rotation is None else torch.as_tensor(rotation)
    if rotation.ndim == 0 or rotation.ndim != 0 and rotation.shape[0] != 3:
        raise ValueError(f'rotation must be of shape (3,)')

    # check scaling factor
    scale = torch.ones(3) if scale is None else torch.as_tensor(scale)
    if scale.ndim == 0:
        scale = scale.repeat(3)
    if scale.shape[0] != 3:
        raise ValueError(f'scale must be of size (3,)')

    # check shearing
    shear = torch.zeros(3) if shear is None else torch.as_tensor(shear)
    if shear.ndim == 0:
        shear = shear.view(1)
    if shear.shape[0] != 3:
        raise ValueError(f'shear must be of shape (3,)')

    # start from translation
    T = torch.eye(4, dtype=torch.float64)
    T[:3, -1] = translation

    # rotation matrix
    R = torch.eye(4, dtype=torch.float64)
    R[:3, :3] = angles_to_rotation_matrix(rotation, degrees=degrees)[:3, :3]

    # scaling
    Z = torch.diag(torch.cat([scale, torch.ones(1, dtype=torch.float64)]))

    # shear matrix
    S = torch.eye(4, dtype=torch.float64)
    S[0][1] = shear[0]
    S[0][2] = shear[1]
    S[1][2] = shear[2]

    # compose component matrices
    matrix = T @ R @ Z @ S

    return AffineMatrix(torch.as_tensor(matrix, dtype=torch.float32, device=device))


def random_affine(
    max_translation: float = 0,
    max_rotation: float = 0,
    max_scaling: float = 0,
    device: torch.device = None) -> AffineMatrix:
    """
    Generate a random affine transformation matrix.

    Args:
        max_translation (float, optional): Maximum translation in each direction.
        max_rotation (float, optional): Maximum rotation in each direction.
        max_scaling (float, optional): Maximum fractional scaling in each direction.
        device (device, optional): Device of the generated affine matrix.

    Returns:
        AffineMatrix: Random affine matrix.
    """
    translation = None
    if max_translation < 0:
        raise ValueError('max_translation must be a positive value')
    if max_translation > 0:
        translation_range = sorted([-max_translation, max_translation])
        translation = torch.distributions.uniform.Uniform(*translation_range).sample((3,))

    rotation = None
    if max_rotation < 0:
        raise ValueError('max_rotation must be a positive value')
    if max_rotation > 0:
        rotation_range = sorted([-max_rotation, max_rotation])
        rotation = torch.distributions.uniform.Uniform(*rotation_range).sample((3,))

    scale = None
    if max_scaling < 0:
        raise ValueError('max_scaling must be a positive value')
    if max_scaling > 0:
        scale = (1 + torch.rand(3) * max_scaling) ** torch.randn(3).sign()

    aff = compose_affine(
        translation=translation,
        rotation=rotation,
        scale=scale,
        device=device)
    return aff


def least_squares_alignment(
    source: torch.Tensor | vx.Mesh,
    target: torch.Tensor | vx.Mesh,
    weights: torch.Tensor = None,
    regularization: float = 1e-6) -> AffineMatrix:
    """
    Compute an affine least squares alignment between two 3D point sets.

    Args:
        source (Tensor or Mesh): Source point set.
        target (Tensor or Mesh): Target point set.
        weights (Tensor, optional): Weights for each point in the source set.
        regularization (float, optional): Regularization scale. Default is 1e-6.
    
    Returns:
        AffineMatrix: Affine alignment matrix.
    """
    if isinstance(source, vx.Mesh):
        source = source.vertices
    if isinstance(target, vx.Mesh):
        target = target.vertices

    # check input shapes
    assert source.shape == target.shape, 'source and target points must have the same shape'
    assert source.shape[1] == 3, 'source and target must be 3D point sets'

    # configure the weight matrix
    if weights is not None:
        assert weights.shape[0] == source.shape[0], 'weights must match the number of points'
        # TODO: ensure that weights are positive
        W = torch.diag(weights)
    else:
        W = torch.eye(len(source), device=source.device, dtype=source.dtype)

    #  extend source to shape (N, 4)
    source = torch.cat([source, torch.ones(source.shape[0], 1, device=source.device)], dim=1)

    # init regularization matrix
    R = regularization * torch.eye(4, device=source.device, dtype=source.dtype)

    # compute weighted least squared estimator
    M = (source.T @ W @ source + R).inverse() @ (source.T @ W @ target)

    return vx.AffineMatrix(M.T)



"""
Reading and writing affines to various file formats.
"""

import os
import torch
import voxel as vx

from .utility import IOProtocol


def load_affine(filename: os.PathLike, fmt: str = None) -> vx.AffineMatrix:
    """
    Load an affine matrix from a file.

    Args:
        filename (PathLike): The path to the file to load.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.
    
    Returns:
        AffineMatrix: The loaded affine matrix.
    """
    vx.io.utility.check_file_readability(filename)

    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(affine_io_protocols, filename)
        if proto is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        proto = vx.io.protocol.find_protocol_by_name(affine_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')

    return proto().load(filename)


def save_affine(affine: vx.AffineMatrix, filename: os.PathLike, fmt: str = None, **kwargs) -> None:
    """
    Save an affine matrix to a file.

    Args:
        affine (AffineMatrix): The affine matrix to save.
        filename (PathLike): The path to the file to save.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.
        kwargs: Additional arguments to pass to the file writing method.
    """
    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(affine_io_protocols, filename)
        if proto is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        proto = vx.io.protocol.find_protocol_by_name(affine_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = proto.enforce_extension(filename)

    proto().save(affine, filename, **kwargs)


class PytorchMatrixIO(IOProtocol):
    """
    IO protocol for storing a simple affine matrix in a pytorch file.
    """
    name = 'torch'
    extensions = ('.pth', '.pt')

    def load(self, filename: os.PathLike) -> vx.AffineMatrix:
        return vx.AffineMatrix(torch.load(filename, weights_only=False))

    def save(self, affine: vx.AffineMatrix, filename: os.PathLike) -> None:
        torch.save(affine.tensor.detach().cpu(), filename)


affine_io_protocols = [
    PytorchMatrixIO,
]



"""
Reading and writing meshes to various file formats.
"""

import os
import numpy as np
import torch
import voxel as vx

from .utility import IOProtocol


def load_mesh(filename: os.PathLike, fmt: str = None, **kwargs) -> vx.Mesh:
    """
    Load a mesh from a file.

    Args:
        filename (PathLike): The path to the file to load.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.
        **kwargs: Additional arguments to pass to the file reading method.

    Returns:
        Mesh: The loaded mesh.
    """
    vx.io.utility.check_file_readability(filename)

    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(mesh_io_protocols, filename)
        if proto is None:
            proto = FreesurferIO
    else:
        proto = vx.io.protocol.find_protocol_by_name(mesh_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')

    return proto().load(filename, **kwargs)


def save_mesh(mesh: vx.Mesh, filename: os.PathLike, fmt: str = None, **kwargs) -> None:
    """
    Save a mesh to a file.

    Args:
        mesh (Mesh): The mesh to save.
        filename (PathLike): The path to the file to save.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.
        kwargs: Additional arguments to pass to the file writing method.
    """
    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(mesh_io_protocols, filename)
        if proto is None:
            proto = FreesurferIO
    else:
        proto = vx.io.protocol.find_protocol_by_name(mesh_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = proto.enforce_extension(filename)

    proto().save(mesh, filename, **kwargs)


class WavefrontIO(IOProtocol):
    """
    Mesh IO protocol for wavefront obj files.
    """

    name = 'obj'
    extensions = ('.obj',)

    def __init__(self):
        try:
            import trimesh
        except ImportError:
            raise ImportError('the trimesh python package must be installed for '
                              'wavefront surface IO')
        self.trimesh = trimesh

    def load(self, filename: os.PathLike) -> vx.Mesh:
        """
        Read mesh from a wavefront object file.

        Args:
            filename (PathLike): The path to the wavefront file to read.

        Returns:
            Mesh: The loaded mesh.
        """
        tmesh = self.trimesh.exchange.load.load(filename, process=False)
        return vx.Mesh(torch.as_tensor(tmesh.vertices), torch.as_tensor(tmesh.faces))

    def save(self, mesh: vx.Mesh, filename: os.PathLike) -> None:
        """
        Write mesh to a wavefront object file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
        """
        parameters = dict(
            header='SPACE=RAS\n',
            include_color=False,
            include_texture=False,
            return_texture=False,
            write_texture=True)
        mesh = mesh.cpu()
        tmesh = self.trimesh.Trimesh(mesh.vertices.detach().numpy(),
                                     mesh.faces.detach().numpy(), process=False)
        tmesh.export(filename, **parameters)


class StanfordPolygonIO(IOProtocol):
    """
    Mesh IO protocol for the Polygon File Format, a.k.a. Stanford PLY.
    """

    name = 'ply'
    extensions = ('.ply',)

    def __init__(self):
        try:
            import trimesh
        except ImportError:
            raise ImportError('the trimesh python package must be installed '
                              'for polygon surface IO')
        self.trimesh = trimesh

    def load(self, filename: os.PathLike) -> vx.Mesh:
        """
        Read mesh from a polygon file.

        Args:
            filename (PathLike): The path to the wavefront file to read.

        Returns:
            Mesh: The loaded mesh.
        """
        tmesh = self.trimesh.exchange.load.load(filename, process=False)
        return vx.Mesh(torch.as_tensor(tmesh.vertices), torch.as_tensor(tmesh.faces))

    def save(self,
        mesh: vx.Mesh,
        filename: os.PathLike,
        vertex_attributes: torch.Tensor = None,
        face_attributes: torch.Tensor = None) -> None:
        """
        Write mesh to a polygon file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
            vertex_attributes (Tensor, optional): Dictionary of vertex attribute
                tensors to save with the mesh.
            face_attributes (Tensor, optional): Dictionary of face attribute tensors
                to save with the mesh.
        """
        def prep_attributes(attributes):
            if not isinstance(attributes, dict):
                attributes = {'attribute': attributes}
            return {k: v.detach().cpu().numpy() for k, v in attributes.items()}

        mesh = mesh.cpu()
        tmesh = self.trimesh.Trimesh(mesh.vertices.detach().numpy(),
                                     mesh.faces.detach().numpy(), process=False)
        if vertex_attributes is not None:
            tmesh.vertex_attributes = prep_attributes(vertex_attributes)
        if face_attributes is not None:
            tmesh.face_attributes = prep_attributes(face_attributes)
        tmesh.export(filename, include_attributes=True)


class TorchMeshIO(IOProtocol):
    """
    Mesh IO protocol for pytorch format.
    """

    name = 'pth'
    extensions = ('.pth', '.pt')

    def load(self, filename: os.PathLike) -> vx.Mesh:
        """
        Read mesh from a file.

        Args:
            filename (PathLike): The path to the file to read.

        Returns:
            Mesh: The loaded mesh.
        """
        items = torch.load(filename, weights_only=False)
        if 'v' not in items or 'f' not in items:
            raise RuntimeError(f'could not find `v` or `f` data keys in {filename}')
        return vx.Mesh(items['v'], items['f'])

    def save(self, mesh: vx.Mesh, filename: os.PathLike) -> None:
        """
        Write mesh to a file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
        """
        vertices = mesh.vertices.detach().cpu()
        faces = mesh.faces.detach().cpu()
        torch.save({'v': vertices, 'f': faces}, filename)


class AbstractTrimeshIO(IOProtocol):
    """
    Mesh IO protocol for the Polygon File Format, a.k.a. Stanford PLY.
    """
    def __init__(self):
        try:
            import trimesh
        except ImportError:
            raise ImportError('the trimesh python package must be installed '
                              f'for {self.name} surface IO')
        self.trimesh = trimesh

    def load(self, filename: os.PathLike) -> vx.Mesh:
        """
        Read mesh from a file.

        Args:
            filename (PathLike): The path to the file to read.

        Returns:
            Mesh: The loaded mesh.
        """
        tmesh = self.trimesh.exchange.load.load(filename, process=False)
        return vx.Mesh(torch.as_tensor(tmesh.vertices), torch.as_tensor(tmesh.faces))

    def save(self, mesh: vx.Mesh, filename: os.PathLike) -> None:
        """
        Write mesh to a file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
        """
        mesh = mesh.cpu()
        tmesh = self.trimesh.Trimesh(mesh.vertices.detach().numpy(),
                                     mesh.faces.detach().numpy(), process=False)
        tmesh.export(filename)


class GltfIO(AbstractTrimeshIO):
    name = 'gltf'
    extensions = ('.gltf', '.glb')


class Stl3D(AbstractTrimeshIO):
    name = 'stl'
    extensions = ('.stl',)


class ThreeDxmlIO(AbstractTrimeshIO):
    name = '3dxml'
    extensions = ('.3dxml',)


class ThreeMfIO(AbstractTrimeshIO):
    name = '3mf'
    extensions = ('.3mf',)


class DrawingExchangeIO(AbstractTrimeshIO):
    name = 'dxf'
    extensions = ('.dxf',)


class ObjectIO(AbstractTrimeshIO):
    name = 'off'
    extensions = ('.off',)


class FreesurferIO(IOProtocol):
    """
    Mesh IO protocol for freesurfer surface files.
    """

    name = 'fs'
    extensions = ('.surf', '.srf', '.fs')

    def __init__(self):
        try:
            import surfa
        except ImportError:
            raise ImportError('the surfa python package must be installed for '
                              'freesurfer surface IO')
        self.surfa = surfa

    def load(self, filename: os.PathLike, space: vx.Space = 'world') -> vx.Mesh:
        """
        Read mesh from a freesurfer surface file.

        Args:
            filename (PathLike): The path to the file to read.
            space (Space): The coordinate space to load the mesh in.
                Can be voxel, surface, or world (default).

        Returns:
            Mesh: The loaded mesh.
        """
        smesh = self.surfa.load_mesh(filename).convert(space=space)
        return vx.Mesh(torch.as_tensor(smesh.vertices), torch.as_tensor(smesh.faces))

    def save(self, mesh: vx.Mesh, filename: os.PathLike) -> None:
        """
        Write mesh to a freesurfer surface file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
        """
        mesh = mesh.cpu()
        smesh = self.surfa.Mesh(mesh.vertices.detach().numpy(),
                                mesh.faces.detach().numpy(), space='world')
        smesh.save(filename)


class GiftiIO(IOProtocol):
    """
    Mesh IO protocol for gifti mesh files.
    """

    name = 'gifti'
    extensions = ('.gii', '.gii.gz')

    def __init__(self):
        try:
            import nibabel
        except ImportError:
            raise ImportError('the nibabel python package must be installed for '
                              'gifti surface IO')
        self.nib = nibabel

    def load(self, filename: os.PathLike) -> vx.Mesh:
        """
        Read mesh from a gifti surface file.

        Args:
            filename (PathLike): The path to the file to read.

        Returns:
            Mesh: The loaded mesh.
        """
        gii = self.nib.load(filename)

        # prefer explicit intents
        verts_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_POINTSET')
        faces_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_TRIANGLE')

        # fall back to guessing based on shape and dtype
        if not verts_arrays or not faces_arrays:
            for da in gii.darrays:
                arr = np.asarray(da.data)
                if arr.ndim == 2 and arr.shape[1] == 3:
                    if not verts_arrays and np.issubdtype(arr.dtype, np.floating):
                        verts_arrays = [da]
                    elif not faces_arrays and np.issubdtype(arr.dtype, np.integer):
                        faces_arrays = [da]

        # if still not found, error out
        if not verts_arrays:
            raise RuntimeError(f'could not find vertex data in {filename}')
        if not faces_arrays:
            raise RuntimeError(f'could not find face data in {filename}')

        # select the first found arrays
        verts = torch.as_tensor(verts_arrays[0].data)
        faces = torch.as_tensor(faces_arrays[0].data)
        return vx.Mesh(verts, faces)

    def save(self, mesh: vx.Mesh, filename: os.PathLike) -> None:
        """
        Write mesh to a gifti surface file.

        Args:
            mesh (Mesh): The mesh to save.
            filename (PathLike): Output path.
        """
        mesh = mesh.cpu()
        verts = mesh.vertices.detach().numpy().astype('float32')
        faces = mesh.faces.detach().numpy().astype('int32')

        intent_codes = self.nib.nifti1.intent_codes

        gda_verts = self.nib.gifti.GiftiDataArray(verts, intent=intent_codes['NIFTI_INTENT_POINTSET'])
        gda_faces = self.nib.gifti.GiftiDataArray(faces,  intent=intent_codes['NIFTI_INTENT_TRIANGLE'])

        gii = self.nib.gifti.GiftiImage(darrays=[gda_verts, gda_faces])
        self.nib.save(gii, filename)


mesh_io_protocols = [
    WavefrontIO,
    StanfordPolygonIO,
    TorchMeshIO,
    GltfIO,
    Stl3D,
    ThreeDxmlIO,
    ThreeMfIO,
    DrawingExchangeIO,
    FreesurferIO,
    GiftiIO,
]



"""
Class utilities for caching parameters computed on-the-fly.
"""

from __future__ import annotations

import functools


def cached(func: callable) -> property:
    """
    Decorator that converts a method into a cached property stored in
    `self._property_cache`. The property is read-only.
    """
    @functools.wraps(func)
    def wrapper(self):
        if func.__name__ not in self._property_cache:
            self._property_cache[func.__name__] = func(self)
        return self._property_cache[func.__name__]
    
    # make the property read-only
    wrapper = property(wrapper)
    return wrapper


def cached_transferable(func: callable) -> property:
    """
    Decorator that converts a method into a cached property stored in
    `self._transferable_property_cache`. The property is read-only.
    A transferable property is a property that can be easily transferred
    to a new instance of the same class.
    """
    @functools.wraps(func)
    def wrapper(self):
        if func.__name__ not in self._property_cache:
            self._transferable_property_cache[func.__name__] = func(self)
        return self._transferable_property_cache[func.__name__]

    # make the property read-only
    wrapper = property(wrapper)
    return wrapper


def init_property_cache(obj: object) -> None:
    """
    Initializes the property cache for an object with functions using
    the `cached` decorator.
    """
    if not hasattr(obj, '_property_cache'):
        obj._property_cache = {}
    if not hasattr(obj, '_transferable_property_cache'):
        obj._transferable_property_cache = {}


def transfer_property_cache(source: object, target: object) -> None:
    """
    Copy the transferable property cache from one object to another.
    """
    if hasattr(source, '_transferable_property_cache'):
        target._transferable_property_cache = source._transferable_property_cache.copy()


def empty_property_cache(obj: object) -> None:
    """
    Clears the entire property cache for an object with
    function using the `cached` decorator.
    """
    if hasattr(obj, '_property_cache'):
        obj._property_cache.clear()
    if hasattr(obj, '_transferable_property_cache'):
        obj._transferable_property_cache.clear()



import torch
import voxel as vx

from . import utility


# TODO: add a mult-frame case to all of these tests


def test_resample():
    
    # resampling to the same resolution should not change the underlying data
    vol = utility.brain_t1w()
    assert vol.tensor is vol.resample_like(vol.geometry).tensor

    # simple test to make sure linear and nearest interpolation are different
    assert not vx.volumes_equal(vol.resample(2, mode='linear'),
                                vol.resample(2, mode='nearest').float())

    # shifted volume should be identical to a rolled volume
    resampled = vol.resample_like(vol.geometry.shift((1, 0, 0), 'voxel'))
    rolled = resampled.new(vol.tensor.roll(-1, 1))
    assert vx.volumes_equal(rolled[:, :-1], resampled[:, :-1])

    # 2x trilinear resampling should be (nearly) identical to average pooling
    resampled = vol.resample(2)
    pool = torch.nn.functional.avg_pool3d
    pooled = resampled.new(pool(vol.tensor.unsqueeze(0).float(), 2, 2).squeeze(0))
    assert vx.volumes_equal(resampled, pooled, vol_tol=3e-3)


def test_antialiasing():

    # verify that antialiasing works as expected by assuming the result
    # is within a certain (manually computed) bound of the non-antialiased result.
    # use source and target volumes with different non-isotropic resolutions and orientations
    vol = utility.brain_t1w().reorient('LIA').resample((0.8, 0.9, 0.8))
    target = vol.geometry.reorient('SPL').resample((1.5, 1.5, 5)).rotate((-10, 15, 10), 'world')
    aa = vol.resample_like(target, antialias=True)
    noaa = vol.resample_like(target, antialias=False)
    error = (aa - noaa).abs().quantile(0.99)
    assert error > 35 and error < 36

    # make sure smoothing is not applied in situations where it should not be
    # i.e. when the downsampling factor is less than 2
    vol = utility.brain_t1w().reorient('IRP').resample((1.8, 1.8, 2))
    target = vol.geometry.reorient('ALS').resample((2.5, 0.6, 0.6))
    aa = vol.resample_like(target, antialias=True)
    noaa = vol.resample_like(target, antialias=False)
    assert vx.volumes_equal(aa, noaa)


def test_point_sampling():

    # make sure that sampling a volume at its own grid points and reshaping
    # the result back to the original shape gives the original volume
    vol = utility.brain_t1w()
    points = vx.volume.volume_grid(vol.baseshape).view(-1, 3)
    sampled = vol.sample(points, space='voxel', mode='nearest').swapaxes(0, 1)
    reshaped = vol.new(sampled.view(vol.shape))
    assert (reshaped == vol).all()
